'' =================================================================================================
''
''   File....... jm_mcp2515.spin
''   Purpose.... MCP2515 CAN interface
''   Author..... Jon "JonnyMac" McPhalen 
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started....  
''   Updated.... 11 DEC 2011 
''
'' =================================================================================================


con

  ' MCP2515 SPI commands

  CMD_RESET      = %1100_0000
  CMD_READ       = %0000_0011                            
  CMD_RD_RXBUF   = %1001_0000                                   ' %1001_0nm0
  CMD_WRITE      = %0000_0010                             
  CMD_LD_TXBUF   = %0100_0000                                   ' %0100_0abc
  CMD_RTS        = %1000_0000                                   ' %1000_0nnn
  CMD_RTS_TXB0   = %1000_0001
  CMD_RTS_TXB1   = %1000_0010
  CMD_RTS_TXB2   = %1000_0100
  CMD_RD_STATUS  = %1010_0000                            
  CMD_RX_STATUS  = %1011_0000
  CMD_BIT_MOD    = %0000_0101


  ' MCP2515 mode (REQOP) bits

  REQOP_CONFIG   = %100_00000 
  REQOP_LISTEN   = %011_00000
  REQOP_LOOPBACK = %010_00000 
  REQOP_SLEEP    = %001_00000 
  REQOP_NORMAL   = $000_00000

  REQOP_MASK     = %111_00000


  #0, PRI_LOW, PRI_MEDLOW, PRI_MEDHI, PRI_HIGH                  ' tx priority settings


con

  #0, PULLUP, DRIVEN
  

var

  long  cs                                                      ' chip select pin (mode == ICS)
  long  sck                                                     ' clock pin
  long  mosi                                                    ' data in
  long  miso                                                    ' data out
  long  lockid                                                  ' 0-7 when sharing resources

  long  inuse
  long  csmode


pub start(cspin, sckpin, mosipin, misopin, lock) 

'' Setup IO for MCP2515
'' -- CS\ pin should be pulled high through 10-47K

  if (inuse)
    stop                                                        ' clear previous definition

  longmove(@cs, @cspin, 5)                                      ' copy parameters

  if (cs < 0)                                                   ' open-drain mode?
    csmode := PULLUP
    cs := ||cs                                                  ' make positive for pin use
    
  else
    csmode := DRIVEN

  chip_select(false)                                            ' deselect

  outa[sck] := 0                                                ' sck low                                                 
  dira[sck] := 1

  outa[mosi] := 0                                               ' mosi low 
  dira[mosi] := 1                                                                                             

  dira[miso] := 0                                               ' miso is input 

  inuse := true
  

pub stop

'' Releases output pins
'' -- use caution if buss pins are shared

  dira[cs] := 0                                                 ' release pins
  dira[sck] := 0
  dira[mosi] := 0

  inuse := false


con

  ' mid/high level commands
  

pub set_mode(mode) | check

  bit_modify(CANCTRL, REQOP_MASK, mode)   


pub get_mode

  return read(CANSTAT) & REQOP_MASK                                     ' return mode bits

  
pub set_baud(kbps) | mode                                               ' section 5.0 of docs

'' Set baud for 16MHz FOsc
'' -- http://www.kvaser.com/en/support/bit-timing-calculator.html
'' -- manually added SAM bit to CNF2
'' -- mode must be set to CONFIG to change baud

  case kbps
    10:
      write(CNF1, $1F)                                                         
      write(CNF2, $FF)                              
      write(CNF3, $07)                                  
      return 10

    20:
      write(CNF1, $0F)                                                         
      write(CNF2, $FF)                              
      write(CNF3, $07)                                    
      return 20

    50:
      write(CNF1, $07)                                                         
      write(CNF2, $EC)                              
      write(CNF3, $07)                                    
      return 50

    100:
      write(CNF1, $03)                                                         
      write(CNF2, $EC)                               
      write(CNF3, $07)                                    
      return 100

    125:
      write(CNF1, $03)                                                         
      write(CNF2, $DA)                              
      write(CNF3, $07)                                    
      return 125

    other:
      return -1


pub tx_buf_ready(buf) | ctrl

'' Returns true if selected buffer (0..2) is available

  if ((buf => 0) and (buf =< 2))
    ctrl := read(TXB0CTRL+buf)
    return ((ctrl & %0000_1000) == 0)  
  else
    return false


pub tx_stdmsg(buf, stdid, dlc, pntr, priority)

'' Send message using 11-bit CAN id
'' -- stdid is the 11-bit msg id
'' -- dlc is the data length control byte (0..8)
'' -- pntr is pointer to byte array with message elements

  case buf
    0:
      write(TXB0SIDH, stdid >> 3)                               ' high bits of id
      write(TXB0SIDL, stdid << 5)                               ' low bits of id
      write(TXB0EID8, 0)                                        ' clear extended id bits
      write(TXB0EID0, 0)
      write(TXB0DLC, dlc)                                       ' set bytes in message
      if (dlc <#= 8)                                            ' any data?
        write_block(TXB0D0, pntr, dlc)                          '  yes, copy data bytes
      write(TXB0CTRL, %1000 | (priority & %11))                 ' send msg with priority 

    1:
      write(TXB1SIDH, stdid >> 3)                           
      write(TXB1SIDL, stdid << 5)  
      write(TXB1EID8, 0)                                         
      write(TXB1EID0, 0)
      write(TXB1DLC, dlc)
      if (dlc <#= 8) 
        write_block(TXB1D0, pntr, dlc)
      write(TXB1CTRL, %1000 | (priority & %11))  

    2:
      write(TXB2SIDH, stdid >> 3)                            
      write(TXB2SIDL, stdid << 5)  
      write(TXB2EID8, 0)                                         
      write(TXB2EID0, 0)
      write(TXB2DLC, dlc)
      if (dlc <#= 8) 
        write_block(TXB2D0, pntr, dlc)
      write(TXB2CTRL, %1000 | (priority & %11)) 
    

con

pub set_rxb0_filters(mask0, filter0, filter1)

'' Sets RXB0 mask and filters

  write(RXM0SIDH, mask0 >> 3)                                 
  write(RXM0SIDL, mask0 << 5)

  write(RXF0SIDH, filter0 >> 3)                               
  write(RXF0SIDL, filter0 << 5)

  write(RXF1SIDH, filter1 >> 3)                               
  write(RXF1SIDL, filter1 << 5)  


pub set_rxb1_filters(mask1, filter2, filter3, filter4, filter5)

'' Sets RXB1 mask and filters

  write(RXM1SIDH, mask1 >> 3)                              
  write(RXM1SIDL, mask1 << 5)

  write(RXF2SIDH, filter2 >> 3)                            
  write(RXF2SIDL, filter2 << 5)

  write(RXF3SIDH, filter3 >> 3)                            
  write(RXF3SIDL, filter3 << 5)  

  write(RXF4SIDH, filter4 >> 3)                            
  write(RXF4SIDL, filter4 << 5)

  write(RXF5SIDH, filter5 >> 3)                            
  write(RXF5SIDL, filter5 << 5)


pub rx_stdmsg(buf, msgpntr, dlcpntr, datapntr) | idx

'' Returns true when std msg retrieved from buf
'' -- msgpntr is pointer to word for can msg id (11-bit)
'' -- dlcpntr is pointer to byte for dlc (data length control)
'' -- datapntr is pointer to array of 8 bytes for msg data

  case buf
    0:
      if (rd_status & %01)                                      ' msg in RB0?
        word[msgpntr] := get_rx_stdid(0)                        ' get id
        byte[dlcpntr] := get_rx_dlc(0)                          ' get dlc
        bytefill(datapntr, 0, 8)                                ' clear data buffer
        read_block(RXB0D0, datapntr, byte[dlcpntr])             ' read new data
        return true                                             ' alert caller
          
    1:                             
      if (rd_status & %10 )                                            
        word[msgpntr] := get_rx_stdid(1)                    
        byte[dlcpntr] := get_rx_dlc(1) 
        bytefill(datapntr, 0, 8)
        read_block(RXB1D0, datapntr, byte[dlcpntr])
        return true              
          
  return false                                                  ' no msg in buf                                             


pub get_rx_stdid(buf)

'' Returns standard (11-bit) message id for buf (0 or 1)

  case buf
    0:
      return ((read(RXB0SIDH) << 3) | (read(RXB0SIDL) >> 5)) & $7FF

    1:
      return ((read(RXB1SIDH) << 3) | (read(RXB1SIDL) >> 5)) & $7FF 

    other:
      return -1


pub get_rx_dlc(buf)

'' Returns DLC (data length control) for buf (0 or 1)

  case buf
    0:
      return read(RXB0DLC) 

    1:
      return read(RXB1DLC)

    other:
      return -1


pub clear_rx_intf(buf)

'' Clears interrupt flag bit for buf (0 or 1)
'' -- used when message has be retrieved or processed to release buffer

  case buf
    0:
      bit_modify(CANINTF, %01, 0)
      return 0  

    1:
      bit_modify(CANINTF, %10, 0)
      return 1

    other:
      return -1

 
con

  ' MCP2515 SPI commands                                        ' DS21801 docs section

 
pub reset                                                       ' 12.2

'' Reset the MCP2515
'' -- use when external RC reset not in place
'' -- okay to use with external reset circuit

  if (lockid => 0)                                              ' sharing buss?
    repeat until not lockset(lockid)                            ' - wait for it to be free

  chip_select(true)   
  spi_readwrite(CMD_RESET) 
  chip_select(false) 

  if (lockid => 0)                                              ' release buss to other cogs
    lockclr(lockid)                                          
      

pub read(addr) | dataout                                        ' 12.3

'' Reads one byte from address in MCP2515

  if (lockid => 0)                                          
    repeat until not lockset(lockid)                        

  chip_select(true) 
  spi_readwrite(CMD_READ)                                     
  spi_readwrite(addr)                                         
  dataout := spi_readwrite(0)                                 
  chip_select(false)

  if (lockid => 0)
    lockclr(lockid) 

  return dataout


pub read_block(addr, pntr, n)

'' Reads block of n bytes from address in MCP2515
'' -- destination is array at pntr

  if (n =< 0)                                                   ' nothing to read?
    return                                                      '  abort

  if (lockid => 0)                                          
    repeat until not lockset(lockid)
  
  chip_select(true) 
  spi_readwrite(CMD_READ)                                    
  spi_readwrite(addr)                                        
  repeat n
    byte[pntr++] := spi_readwrite(0)                         
  chip_select(false)

  if (lockid => 0)
    lockclr(lockid) 


pub read_rxbuf(bufsel) | dataout                                ' 12.4

'' Read RX buffer
'' -- %00 = RXB0SIDH, %01 = RXB0D0
'' -- %10 = RXB1SIDH, %11 = RXB1D0 

  if (lockid => 0)                                          
    repeat until not lockset(lockid)

  chip_select(true) 
  spi_readwrite(CMD_RD_RXBUF | (bufsel << 1))
  dataout := spi_readwrite(0)
  chip_select(false)

  if (lockid => 0)
    lockclr(lockid) 

  return dataout


pub write(addr, value)                                          ' 12.5

'' Write one byte to addr in MCP2515

  if (lockid => 0)                                          
    repeat until not lockset(lockid)

  chip_select(true) 
  spi_readwrite(CMD_WRITE)
  spi_readwrite(addr)
  spi_readwrite(value)
  chip_select(false)

  if (lockid => 0)
    lockclr(lockid) 
  

pub write_block(addr, pntr, n)

'' Write n bytes from pntr (hub address) to addr in MCP2515

  if (lockid => 0)                                          
    repeat until not lockset(lockid)

  chip_select(true) 
  spi_readwrite(CMD_WRITE)
  spi_readwrite(addr)
  repeat n
    spi_readwrite(byte[pntr++])
  chip_select(false)

  if (lockid => 0)
    lockclr(lockid) 

    
pub load_txbuf(datain, bufsel)                                  ' 12.6

'' Load TX buffer
'' -- %000 = TXB0SIDH, %001 = TXB0D0
'' -- %010 = TXB1SIDH, %011 = TXB1D0
'' -- %100 = TXB2SIDH, %101 = TXB2D0

  if (lockid => 0)                                          
    repeat until not lockset(lockid)    

  chip_select(true) 
  spi_readwrite(CMD_LD_TXBUF | (bufsel & %111))
  spi_readwrite(datain)
  chip_select(false)

  if (lockid => 0)
    lockclr(lockid) 

  
pub rts(buf)                                                    ' 12.7

  if (lockid => 0)                                          
    repeat until not lockset(lockid)

  chip_select(true) 
  spi_readwrite(CMD_RTS | (buf & %111))
  chip_select(false)

  if (lockid => 0)
    lockclr(lockid)  


pub rd_status | statusbits                                      ' 12.8

'' Reads status bits from MCP2515

  if (lockid => 0)                                          
    repeat until not lockset(lockid)

  chip_select(true) 
  spi_readwrite(CMD_RD_STATUS)                                
  statusbits := spi_readwrite(0)                              
  chip_select(false)

  if (lockid => 0)
    lockclr(lockid)  

  return statusbits


pub rx_status | statusbits                                      ' 12.9

'' Use to determine filter match for incoming message

  if (lockid => 0)                                          
    repeat until not lockset(lockid)
    
  chip_select(true) 
  spi_readwrite(CMD_RX_STATUS)                                 
  statusbits := spi_readwrite(0)                               
  chip_select(false)

  if (lockid => 0)
    lockclr(lockid) 

  return statusbits

  
pub bit_modify(addr, mask, datain)                              ' 12.10

  if (lockid => 0)                                        
    repeat until not lockset(lockid)                      

  chip_select(true)                                               
  spi_readwrite(CMD_BIT_MOD)
  spi_readwrite(addr)
  spi_readwrite(mask)
  spi_readwrite(datain)
  chip_select(false)

  if (lockid => 0)
    lockclr(lockid)                                                                                     

  
con

pri chip_select(state)

'' Select MCP2515
'' -- pulls CS\ line low when state is true

  case csmode
    DRIVEN:
      if (state)       
        outa[cs] := 0                                           ' drive low  
        dira[cs] := 1       
      else              
        outa[cs] := 1                                           ' drive high  
        dira[cs] := 1 

    PULLUP:
      if (state)
        outa[cs] := 0                                           ' drive low
        dira[cs] := 1
      else
        dira[cs] := 0                                           ' release to pull-up

      
pri spi_readwrite(datain) | dataout

'' Shifts datain into MCP2515, receives and returns dataout (from MCP2515)

  datain <<= constant(32-8)                                     ' prep for MSB output
  repeat 8
    outa[mosi] := (datain <-= 1) & 1                            ' datain msb --> MOSI  
    dataout := (dataout << 1) | ina[miso]                       ' MISO <-- dataout msb 
    outa[sck] := 1                                              ' clock high
    outa[sck] := 0                                              ' clock low
   
  return dataout & $FF
  

con

  ' Control registers

  BFPCTRL       = $0C
  TXRTSCTRL     = $0D  
  CANSTAT       = $0E
  CANCTRL       = $0F
  TEC           = $1C
  REC           = $1D
  CNF3          = $28
  CNF2          = $29
  CNF1          = $2A
  CANINTE       = $2B
  CANINTF       = $2C
  EFLG          = $2D
  TXB0CTRL      = $30  
  TXB1CTRL      = $40  
  TXB2CTRL      = $50
  RXB0CTRL      = $60
  RXB1CTRL      = $70
     

  ' MCP2515 Registers

  RXF0SIDH      = $00
  RXF0SIDL      = $01
  RXF0EID8      = $02
  RXF0EID0      = $03
  RXF1SIDH      = $04
  RXF1SIDL      = $05
  RXF1EID8      = $06
  RXF1EID0      = $07
  RXF2SIDH      = $08
  RXF2SIDL      = $09
  RXF2EID8      = $0A
  RXF2EID0      = $0B
' BFPCTRL       = $0C
' TXRTSCTRL     = $0D
' CANSTAT       = $0E
' CANCTRL       = $0F
   
  RXF3SIDH      = $10
  RXF3SIDL      = $11
  RXF3EID8      = $12
  RXF3EID0      = $13
  RXF4SIDH      = $14
  RXF4SIDL      = $15
  RXF4EID8      = $16
  RXF4EID0      = $17
  RXF5SIDH      = $18
  RXF5SIDL      = $19
  RXF5EID8      = $1A
  RXF5EID0      = $1B
' TEC           = $1C
' REC           = $1D
  CANSTAT1      = $1E
  CANCTRL1      = $1F               
                 
  RXM0SIDH      = $20
  RXM0SIDL      = $21
  RXM0EID8      = $22
  RXM0EID0      = $23
  RXM1SIDH      = $24
  RXM1SIDL      = $25
  RXM1EID8      = $26
  RXM1EID0      = $27
' CNF3          = $28
' CNF2          = $29
' CNF1          = $2A
' CANINTE       = $2B
' CANINTF       = $2C
' EFLG          = $2D
  CANSTAT2      = $2E
  CANCTRL2      = $2F
                 
' TXB0CTRL      = $30
  TXB0SIDH      = $31
  TXB0SIDL      = $32
  TXB0EID8      = $33
  TXB0EID0      = $34
  TXB0DLC       = $35
  TXB0D0        = $36
  TXB0D1        = $37
  TXB0D2        = $38
  TXB0D3        = $39
  TXB0D4        = $3A
  TXB0D5        = $3B
  TXB0D6        = $3C
  TXB0D7        = $3D
  CANSTAT3      = $3E
  CANCTRL3      = $3F
                 
' TXB1CTRL      = $40
  TXB1SIDH      = $41
  TXB1SIDL      = $42
  TXB1EID8      = $43
  TXB1EID0      = $44
  TXB1DLC       = $45
  TXB1D0        = $46
  TXB1D1        = $47
  TXB1D2        = $48
  TXB1D3        = $49
  TXB1D4        = $4A
  TXB1D5        = $4B
  TXB1D6        = $4C
  TXB1D7        = $4D
  CANSTAT4      = $4E
  CANCTRL4      = $4F
                 
' TXB2CTRL      = $50
  TXB2SIDH      = $51
  TXB2SIDL      = $52
  TXB2EID8      = $53
  TXB2EID0      = $54
  TXB2DLC       = $55
  TXB2D0        = $56
  TXB2D1        = $57
  TXB2D2        = $58
  TXB2D3        = $59
  TXB2D4        = $5A
  TXB2D5        = $5B
  TXB2D6        = $5C
  TXB2D7        = $5D
  CANSTAT5      = $5E
  CANCTRL5      = $5F
                 
' RXB0CTRL      = $60
  RXB0SIDH      = $61
  RXB0SIDL      = $62
  RXB0EID8      = $63
  RXB0EID0      = $64
  RXB0DLC       = $65
  RXB0D0        = $66
  RXB0D1        = $67
  RXB0D2        = $68
  RXB0D3        = $69
  RXB0D4        = $6A
  RXB0D5        = $6B
  RXB0D6        = $6C
  RXB0D7        = $6D
  CANSTAT6      = $6E
  CANCTRL6      = $6F
                 
' RXB1CTRL      = $70
  RXB1SIDH      = $71
  RXB1SIDL      = $72
  RXB1EID8      = $73
  RXB1EID0      = $74
  RXB1DLC       = $75
  RXB1D0        = $76
  RXB1D1        = $77
  RXB1D2        = $78
  RXB1D3        = $79
  RXB1D4        = $7A
  RXB1D5        = $7B
  RXB1D6        = $7C
  RXB1D7        = $7D
  CANSTAT7      = $7E
  CANCTRL7      = $7F

  
dat

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}